# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
# Date: 06.09.2016                                                            #
# Author: Ole-Johan Skrede                                                    #
#                                                                             #
# Solution proposal as part of the exercise program in                        #
# INF4300 - Digital image analysis at the University of Oslo                  #
#                                                                             #
# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
"""
Implementation of Hough Transform
"""

# pylint: disable=bad-indentation

import random

import numpy as np
import cv2
import matplotlib.pyplot as plt
from skimage.draw import line_aa # pylint: disable=import-error


class RandomHoughTransform():
  """Random Hough transform class. Currently, only support for lines."""
  def __init__(self, image_file, which_object='line', acc_threshold=100,
               obj_threshold=100):
    """
    Args:
      image_file: string. Full filepath of input image.
      which_object: string. {line}
      acc_threshold: int. Number of votes a point in the accumulator needs to
                     have before we check if it constitute an object.
      obj_threshold: int. Number of pixels in a proposed solution that needs to
                     overlap with foreground pixels in order to verify that the
                     proposed solution is indeed a solution.
    """
    self.image = cv2.imread(image_file, cv2.IMREAD_COLOR)
    self.height = self.image.shape[0]
    self.width = self.image.shape[1]
    self.max_length = np.sqrt(self.height*self.height + self.width*self.width)
    self.gray_image = cv2.cvtColor(self.image, cv2.COLOR_BGR2GRAY)
    self.which_object = which_object
    self.acc_threshold = acc_threshold
    self.obj_threshold = obj_threshold
    self.lined_image = np.copy(self.image)

    # Resolution of parameters. For every parameter, we update the accumulator
    # for this value. Since we are dealing with floats, we say that two
    # parameters are equal if they are closer than a value given by this
    # resolution
    self.res_param1 = 1
    self.res_param2 = np.pi / 180
    self.res_param3 = 1

    # Neighbourhood around parameters.
    self.neigh_param1 = 5*self.res_param1
    self.neigh_param2 = 5*self.res_param2
    self.neigh_param3 = 5*self.res_param3

    self.edges = self.preprocess(self.gray_image, 'canny')

    temp_fg = np.where(self.edges > 0)
    self.foreground = set([(x, y) for x, y in zip(temp_fg[0], temp_fg[1])])
    self.accumulator = [] # list of tuples (param1, param2, ..., count)
    self.blacklist = set() # set of parameter tuples that we have checked, and
                           # does not constitute an interesting object
    self.solutions = set() # set of parameter tuples contituating interesting
                           # objects.

  def preprocess(self, image=None, method='canny'):
    """Preprocess"""
    if image is None:
      image = self.gray_image

    # Illuminate edges
    if method == 'sobel':
      image = cv2.medianBlur(image, 11)
      hor_grad = cv2.Sobel(image, cv2.CV_64F, 1, 0, ksize=5)
      ver_grad = cv2.Sobel(image, cv2.CV_64F, 0, 1, ksize=5)
      grad_magn = np.sqrt(hor_grad*hor_grad + ver_grad*ver_grad)
      grad_magn = ((grad_magn - np.min(grad_magn)) /
                   (np.max(grad_magn) - np.min(grad_magn))*255)
      edges = (grad_magn <= 200)*0 + (grad_magn > 200)*255
    elif method == 'canny':
      image = cv2.medianBlur(image, 5)
      edges = cv2.Canny(image,
                        100, # First hysteresis threshold
                        150, # Second hysteresis threshold,
                        apertureSize=3)


    return edges

  @staticmethod
  def points_to_parameters_line(point1, point2):
    """
    Use the two points in image space (x,y), to create two parameter values
    (rho, theta), by solving the system

      rho = x1*cos(theta) + y1*sin(theta)
      rho = x2*cos(theta) + y2*sin(theta)

    Args:
      point* (int, int). (x*, y*), (row_index, col_index).

    Returns:
      rho: float. Distance (in pixel units).
      theta: float. Angle (in radians).
    """
    x1, y1 = point1
    x2, y2 = point2
    if y1 != y2:
      theta = np.arctan((x1 - x2) / (y2 - y1))
    else:
      if x1 > x2:
        theta = np.pi / 2
      else:
        theta = -np.pi / 2
    rho = x1*np.cos(theta) + y1*np.sin(theta)
    test_rho = x2*np.cos(theta) + y2*np.sin(theta)
    assert np.abs(rho - test_rho) < 1e-6, 'Something fishy with line equation'

    return rho, theta

  def create_candidates(self, rho, theta):
    """Create candidate parameters based on accumulator value"""
    exist_in_accumulator = False
    candidate_solution = False
    for ind, point in enumerate(self.accumulator):
      # Check if point is already there, and increment the count if it is
      ex_rho = point[0]
      ex_theta = point[1]
      if ((np.abs(ex_rho - rho) < self.res_param1) and
              (np.abs(ex_theta - theta) < self.res_param2)):
        exist_in_accumulator = True
        cand_rho = ex_rho
        cand_theta = ex_theta
        self.accumulator[ind][2] += 1
        num_votes = self.accumulator[ind][2]
        if num_votes > self.acc_threshold:
          candidate_solution = True
        break

    if not exist_in_accumulator:
      # Check if there is a point close to it, if it is, find the average between
      # the two, and update the accumulator to this average (also increment the
      # count)
      for ind, point in enumerate(self.accumulator):
        ex_rho = point[0]
        ex_theta = point[1]
        if ((np.abs(ex_rho - rho) < self.neigh_param1) and
                (np.abs(ex_theta - theta) < self.neigh_param2)):
          exist_in_accumulator = True
          cand_rho = (ex_rho + rho) / 2
          cand_theta = (ex_theta + theta) / 2
          self.accumulator[ind][0] = cand_rho
          self.accumulator[ind][1] = cand_theta
          self.accumulator[ind][2] += 1
          num_votes = self.accumulator[ind][2]
          if num_votes > self.acc_threshold:
            candidate_solution = True
          break

    if not exist_in_accumulator:
      # Create a new entry for this point
      self.accumulator.append([rho, theta, 1])
      cand_rho = rho
      cand_theta = theta
      num_votes = 1
      if num_votes > self.acc_threshold: # Here count is only 1
        candidate_solution = True

    return cand_rho, cand_theta, candidate_solution

  def check_solution(self, rho, theta):
    """Check if candidate solution is indeed a solution"""

    # Create line image based on rho and theta
    cost = np.cos(theta)
    sint = np.sin(theta)
    x0 = rho*cost
    y0 = rho*sint
    x1 = int(x0 + self.max_length*(-sint))
    y1 = int(y0 + self.max_length*(cost))
    x2 = int(x0 - self.max_length*(-sint))
    y2 = int(y0 - self.max_length*(cost))
    line_x, line_y, _ = line_aa(x1, y1, x2, y2) # TODO: Consider thresholding
                                                # it by value (third return)
    # Select only the part of the line inside the image
    line = [(x, y) for x, y in zip(line_x, line_y) if (x >= 0 and y >= 0 and
                                                       x < self.height and
                                                       y < self.width)]

    # Check how many points in the line that are foreground points. If there
    # are enough, delete these overlapping foreground points from
    # self.foreground
    num_intersections = 0
    intersection_points = self.foreground.intersection(line)
    num_intersections = len(intersection_points)
    if num_intersections > self.obj_threshold:
      self.foreground = self.foreground - intersection_points # NB set types
      self.solutions.add((rho, theta))
    else:
      self.blacklist.add((rho, theta))


  def run(self, max_iter):
    """Run the Hough transform."""
    iter_count = 0
    continue_search = True
    while continue_search:
      # Pick two different random points from the foreground
      point1, point2 = random.sample(self.foreground, 2)
      rho, theta = self.points_to_parameters_line(point1, point2)
      if (rho, theta) not in self.blacklist.union(self.solutions):
        rho_cand, theta_cand, cand_solution = self.create_candidates(rho, theta)
        if cand_solution:
          self.check_solution(rho_cand, theta_cand)
      iter_count += 1

      # Check convergence criteria
      if len(self.foreground) == 0:
        continue_search = False
        print('No more foreground pixels, search terminated')
      if iter_count >= max_iter:
        continue_search = False
        print('Max iterations reached, search terminated')

    print('Found {} solutions'.format(len(self.solutions)))

  def draw_solution_image(self):
    """Draw lines in original image"""
    for rho, theta in self.solutions:
      cost = np.cos(theta)
      sint = np.sin(theta)
      x0 = cost*rho
      y0 = sint*rho
      x1 = int(x0 + self.max_length*(-sint))
      y1 = int(y0 + self.max_length*(cost))
      x2 = int(x0 - self.max_length*(-sint))
      y2 = int(y0 - self.max_length*(cost))

      cv2.line(self.lined_image, (y1, x1), (y2, x2), (0, 0, 255), 2)
